#include "StringArray.h"
#include "Error.h"
#include <stdio.h>
#include "math.h"
#include <iostream>
#include "StringHash.h"
#include "MathStats.h"
#include "MixChidist.h"
#include "MathSVD.h"
#include "Meta.h"
#define MATHLIB_STANDALONE 
#include <Rmath.h>
#include "Calculate_mvt_pvalue.h"
#include "My_mvt.h"

String Meta::studyName = "";
String Meta::prefix = ""; 
bool Meta::Burden = false;
bool Meta::MB = false;
bool Meta::SKAT = false;
bool Meta::VTa = false;
bool Meta::VTp = false;
bool Meta::outvcf =false; 
bool Meta::fullResult =false; 
bool Meta::report = false; 
double  Meta::report_pvalue_cutoff = 1e-06; 
bool Meta::founderAF = false; 
double Meta::HWE = 0.0;
double Meta::CALLRATE = 0.0;
double Meta::MAF_cutoff = 0.05;

Meta::Meta(){}
Meta::~Meta()
{}

//This function read all study names from a file 
//and save the names in the StringArray files
void Meta::Prepare()
{
   if(studyName!="")
   {
      FILE * inFile = fopen(studyName,"r");
      if(inFile==NULL)
	 error("FATAL ERROR! Please check file name for --studyName  %s\n",studyName.c_str());
      String buffer;
      while (!feof(inFile))
      {
	 buffer.ReadLine(inFile);
	 if(buffer.FindChar('#')!=-1)
	    continue;
	 studies.AddTokens(buffer, "\n");
      }
      fclose(inFile);
   }
   else
      error("FATAL ERROR! --studyName can not be empty! Usage: --studyName yourStudynames\n");
}

//this function will read through summary statistics of each study
//and pool the information.
void Meta::PoolSummaryStat()
{
   //usefulSize and usefulAC have the pooled N and AC information 
   StringIntHash usefulSize;
   StringDoubleHash usefulAC; 
   //refalt use study:chr:pos as key and ref:alt as value
   StringHash refalt;
   StringIntHash SNP_DirectionByStudy,SNP_DirectionByStudyNoAllele;
   StringArray directions;

   total_N=0;
   int flipCount=0;

   for(int study=0;study<studies.Length();study++)
   {
      //#CHROM  POS     REF     ALT     N_INFORMATIVE   FOUNDER_AF      ALL_AF  INFORMATIVE_ALT_AC      CALL_RATE       HWE_PVALUE      N_REF   N_HET   N_ALT   U_STAT  SQRT_V_STAT     ALT_EFFSIZE     PVALUE

      printf("Pooling summary statistics from study %s ...\n",studies[study].c_str());

      //read in summary statistics. 
      //maf and summary stat are saved. SNPpos in the file are hashed.
      String filename = studies[study] + ".singlevar.score.txt";
      FILE * file;
      file = fopen(filename,"r");
      int adjust=0;

      if(file == NULL)
      {
	 filename = studies[study] + ".MetaScore.assoc";
	 file = fopen(filename,"r");
	 adjust=1;
      }

      if(file == NULL)
	 error("Cannot open file: %s!\n",filename.c_str());

      String buffer;
      StringArray tokens;
      int current_N =0;
      while (!feof(file))
      {
	 buffer.ReadLine(file);
	 //get sample size
	 if(buffer.Find("##AnalyzedSamples")!=-1)
	 {
	    StringArray tokens;
	    tokens.AddTokens(buffer,"=");
	    SampleSize.Push(tokens[1].AsInteger());
	    continue;
	 }
	 else if(buffer.FindChar('#') !=-1 || buffer.Find("CHROM")!=-1)
	    continue;
	 tokens.Clear();
	 tokens.AddTokens(buffer, SEPARATORS);

	 if(tokens[0].Find("chr")!=-1)
	    tokens[0] = tokens[0].SubStr(3);

	 String tmp_SNP = tokens[0]+":"+tokens[1]+":"+tokens[2]+":"+tokens[3];
	 String tmp_SNP_flip = tokens[0]+":"+tokens[1]+":"+tokens[3]+":"+tokens[2];
	 String refalt_current = study + ":" + tokens[0] + ":" + tokens[1];

	 //check if this SNP (allowing allele flips) has been hashed before
	 int index1 = usefulAC.Find(tmp_SNP);
	 int index2 = usefulAC.Find(tmp_SNP_flip);

	 //QC: filtering variants
	 if(tokens[8-adjust].AsDouble()<CALLRATE || tokens[9-adjust].AsDouble()<HWE)
	 {
	    int directions_idx;
	    if(index1 !=-1)
	    {
	       directions_idx = SNP_DirectionByStudy.Integer(tmp_SNP);
	       if(directions[directions_idx].Length()<study)
	       {
		  int len = directions[directions_idx].Length();
		  for(int s=0;s<study+1-len;s++)
		     directions[directions_idx] += "?";
	       }
	       directions[directions_idx] += "!";
	    }
	    else if(index2 != -1)
	    {
	       directions_idx = SNP_DirectionByStudy.Integer(tmp_SNP_flip);
	       if(directions[directions_idx].Length()<study)
	       {
		  int len = directions[directions_idx].Length();
		  for(int s=0;s<study+1-len;s++)
		     directions[directions_idx] += "?";
	       }
	       directions[directions_idx] += "!";
	    }
	    else
	    {
	       String * refalt_saved = (String *) refalt.Object(tokens[0]+":"+tokens[1]);
	       if(refalt_saved != NULL)
	       {
		  //if this variant at this chr:pos has been hashed but with different ref/alt alleles
		  //then skip this variant
		  SNPexclude.SetInteger(study+":"+tokens[0]+":"+tokens[1],0);
		  directions_idx = SNP_DirectionByStudyNoAllele.Integer(tokens[0]+":"+tokens[1]);
		  if(directions[directions_idx].Length()<study)
		  {
		     int len = directions[directions_idx].Length();
		     for(int s=0;s<study+1-len;s++)
			directions[directions_idx] += "?";
		  }
		  directions[directions_idx] += "!";
		  continue;
	       }

	       String direction = "";
	       if(study>0)
	       {
		  for(int s=0;s<study;s++)
		     direction += "?";
	       }
	       direction += "!";
	       directions.Push(direction);
	       SNP_DirectionByStudy.SetInteger(tmp_SNP,directions.Length()-1);
	       SNP_DirectionByStudyNoAllele.SetInteger(tmp_SNP,directions.Length()-1);
	    }
	    continue;
	 }

	 int current_AC;
	 if(tokens[5]=="NA")
	 {
	    current_AC = 0;
	    current_N = SampleSize[study];
	 }
	 else
	 {
	    current_AC = 2*tokens[12-adjust].AsInteger() + tokens[11-adjust].AsInteger();
	    current_N = tokens[10-adjust].AsInteger()+tokens[11-adjust].AsInteger()+tokens[12-adjust].AsInteger();
	 }

	 if(index1 !=-1 || index2 !=-1)
	 {
	    if(index2!=-1)
	    {
	       flipCount++;
	       flipSNP.SetInteger(study+":"+tokens[0]+":"+tokens[1],flipCount);
	    }
	    //update usefulAC
	    //handling flipped alleles
	    if(index2!=-1)	       
	    {
	       double AC = usefulAC.Double(index2);
	       if(founderAF)
	       {
		  AC += 2.0*SampleSize[study]*(1.0-tokens[5].AsDouble());
		  usefulAC.SetDouble(index2,AC);
	       }
	       else
	       {
		  AC += current_N*2.0-current_AC;  
		  usefulAC.SetDouble(index2,AC);
	       }
	    }
	    else 
	    {
	       double AC = usefulAC.Double(index1);
	       if(founderAF)
	       {
		  AC += 2.0*SampleSize[study]*tokens[5].AsDouble();
		  usefulAC.SetDouble(index1,AC);
	       }
	       else
	       {
		  AC += current_AC;
		  usefulAC.SetDouble(index1,AC);
	       }
	    }

	    //update usefulSize
	    int N_idx;
	    if(index1!=-1)
	       N_idx= usefulSize.Find(tmp_SNP);
	    else
	       N_idx= usefulSize.Find(tmp_SNP_flip);

	    int N = usefulSize.GetCount(N_idx); 

	    if(founderAF)
	    {
	       N+=SampleSize[study];
	    }
	    else
	    {
	       N += current_N;
	    }

	    usefulSize.SetInteger(N_idx,N);

	    //update SNPstat
	    int stat_idx;
	    if(index1!=-1)
	    {
	       stat_idx = SNPstat.Find(tmp_SNP);
	       double stat = SNPstat.Double(stat_idx);
	       stat += tokens[13-adjust].AsDouble();
	       SNPstat.SetDouble(stat_idx,stat);
	    }
	    else
	    {
	       stat_idx = SNPstat.Find(tmp_SNP_flip);
	       double stat = SNPstat.Double(stat_idx);
	       stat += -1.0*tokens[13-adjust].AsDouble();
	       SNPstat.SetDouble(stat_idx,stat);
	    }

	    //update SNP_Vstat
	    int Vstat_idx;
	    if(index1!=-1)
	       Vstat_idx = SNP_Vstat.Find(tmp_SNP);
	    else
	       Vstat_idx = SNP_Vstat.Find(tmp_SNP_flip);

	    double Vstat = SNP_Vstat.Double(Vstat_idx);
	    double vstat_tmp = tokens[14-adjust].AsDouble();
	    vstat_tmp = vstat_tmp*vstat_tmp;
	    Vstat += vstat_tmp;
	    SNP_Vstat.SetDouble(Vstat_idx,Vstat);

	    //update SNP_DirectionByStudy
	    int directions_idx;
	    String direction = "+";
	    if(tokens[15-adjust]=="NA" || tokens[15-adjust]=="na" || tokens[15-adjust]=="nan" || tokens[15-adjust]=="-")
	    {
	       direction = "?";
	    }

	    if(index1!=-1)
	    {
	       if(tokens[15-adjust].AsDouble()<0)
		  direction = "-";
	       directions_idx = SNP_DirectionByStudy.Integer(tmp_SNP);	
	    }
	    else
	    {
	       if(tokens[15-adjust].AsDouble()>0)
		  direction = "-";
	       directions_idx = SNP_DirectionByStudy.Integer(tmp_SNP_flip);	
	    }
	    if(directions[directions_idx].Length()<study)
	    {
	       int len = directions[directions_idx].Length();
	       for(int s=0;s<study+1-len;s++)
		  directions[directions_idx] += "?";
	    }
	    directions[directions_idx] += direction;
	 }
	 else
	 {
	    String * refalt_saved = (String *) refalt.Object(tokens[0]+":"+tokens[1]);
	    if(refalt_saved != NULL)
	    {
	       //if this variant at this chr:pos has been hashed but with different ref/alt alleles
	       //then skip this variant
	       SNPexclude.SetInteger(study+":"+tokens[0]+":"+tokens[1],0);
	       int directions_idx = SNP_DirectionByStudyNoAllele.Integer(tokens[0]+":"+tokens[1]);
	       if(directions[directions_idx].Length()<study)
	       {
		  int len = directions[directions_idx].Length();
		  for(int s=0;s<study+1-len;s++)
		     directions[directions_idx] += "?";
	       }
	       directions[directions_idx] += "!";
	       continue;
	    }

	    if(founderAF)
	    {
	       usefulAC.SetDouble(tmp_SNP,2.0*SampleSize[study]*tokens[5].AsDouble());
	       usefulSize.SetInteger(tmp_SNP,SampleSize[study]);
	    }
	    else
	    {
	       usefulAC.SetDouble(tmp_SNP,current_AC);
	       usefulSize.SetInteger(tmp_SNP,current_N);
	    }
	    SNPstat.SetDouble(tmp_SNP,tokens[13-adjust].AsDouble());
	    double vstat_tmp = tokens[14-adjust].AsDouble();
	    vstat_tmp = vstat_tmp*vstat_tmp;
	    SNP_Vstat.SetDouble(tmp_SNP,vstat_tmp);
	    refalt.SetObject(tokens[0]+":"+tokens[1], &refalt_current);


	    String direction = "";
	    if(study!=0)
	    {
	       for(int s=0;s<study;s++)
		  direction += "?";
	    }

	    if(tokens[15-adjust]=="NA" || tokens[15-adjust]=="na" || tokens[15-adjust]=="nan" || tokens[15-adjust]=="-")
	       direction += "?";
	    else if(tokens[15-adjust].AsDouble()<0)
	       direction += "-";
	    else if(tokens[15-adjust].AsDouble()>0)
	       direction += "+";

	    directions.Push(direction);
	    SNP_DirectionByStudy.SetInteger(tmp_SNP,directions.Length()-1);
	    SNP_DirectionByStudyNoAllele.SetInteger(tokens[0]+":"+tokens[1],directions.Length()-1);
	 }
      }
      fclose(file);

      total_N += SampleSize[study];
   }

   //calculate pooled allele frequencies here
   for(int i=0;i<usefulAC.Capacity();i++)
   {
      if(!usefulAC.SlotInUse(i)) 
      {
	 continue;
      }
      String SNPname = usefulAC[i];
      double AC = usefulAC.Double(i);
      int N = usefulSize.Integer(SNPname);
      double maf;
      if(founderAF)
      {
	 maf = AC/(2.0*total_N);
      }
      else
      {
	 maf = AC/(2.0*N);
      }
      if(maf>0.5) 
	 maf = 1.0-maf;

      SNPmaf.SetDouble(SNPname,maf);
   } //now all SNP maf are hashed in SNPmaf.

   //calculate single variant meta results
   printf("Start single variant meta analysis ...\n");
   //calculate final results here
   FILE * output;
   String filename;
   if(prefix =="")
      filename = "meta.singlevar.results";
   else if(prefix.Last()=='.' || prefix.Last()=='/')
      filename = prefix +  "meta.singlevar.results";
   else
      filename = prefix + ".meta.singlevar.results";

   output=fopen(filename,"wt");

   fprintf(output,"##Method=SinglevarScore\n");
   fprintf(output,"##STUDY_NUM=%d\n",studies.Length());
   fprintf(output,"##TotalSampleSize=%d\n",total_N);
   fprintf(output,"#CHROM\tPOS\tREF\tALT\tPOOLED_ALT_AF\tEFFECT_SIZE\tDIRECTION_BY_STUDY\tPVALUE\n");

   FILE * vcfout;
   if(outvcf)
   {
      String filename;
      if(prefix=="")
	 filename = "pooled.variants.vcf";
      else if(prefix.Last()=='.' || prefix.Last()=='/')
	 filename = prefix + "pooled.variants.vcf";
      else 
	 filename = prefix + ".pooled.variants.vcf";

      vcfout = fopen(filename,"wt");
      fprintf(vcfout,"#CHR\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\n");
   }

   for(int i=0;i<SNPmaf.Capacity();i++)
   {
      if(!SNPmaf.SlotInUse(i))
      {
	 continue;
      }
      String SNPname = SNPmaf[i];
      double U = SNPstat.Double(SNPname);
      double V = SNP_Vstat.Double(SNPname);
      double maf = SNPmaf.Double(SNPname);
      int direction_idx = SNP_DirectionByStudy.Integer(SNPname);
      if(directions[direction_idx].Length()<studies.Length())
      {
	 int len = directions[direction_idx].Length();
	 for(int s=0;s<studies.Length()-len;s++)
	    directions[direction_idx] += "?";
      }

      String direction = directions[direction_idx];
      //printf("direction is %s,SNPname is %s\n",directions[direction_idx].c_str(), SNPname.c_str());

      if(maf>0 && maf<1.0 && V!=0.0)
      {
	 double chisq = U*U/V;
	 double pvalue = pchisq(chisq,1,0,0);
	 double effSize = U/V;
	 bool disect=false;
	 while(pvalue==0.0)
	 {
	    disect=true;
	    chisq *= 0.999;
	    pvalue = pchisq(chisq,1,0,0);
	 }
	 StringArray tokens;
	 tokens.AddTokens(SNPname,":");

	 singleVarPvalue.SetDouble(SNPname,pvalue);
	 singleVarEff.SetDouble(SNPname,effSize);

	 if(disect)
	    fprintf(output,"%s\t%s\t%s\t%s\t%g\t%g\t%s\t<%g\n",tokens[0].c_str(),tokens[1].c_str(),tokens[2].c_str(),tokens[3].c_str(),maf,effSize,direction.c_str(),pvalue);
	 else
	    fprintf(output,"%s\t%s\t%s\t%s\t%g\t%g\t%s\t%g\n",tokens[0].c_str(),tokens[1].c_str(),tokens[2].c_str(),tokens[3].c_str(),maf,effSize,direction.c_str(),pvalue);

	 if(outvcf)
	    fprintf(vcfout,"%s\t%s\t%s\t%s\t%s\t.\t.\tALT_AF=%g;\n",tokens[0].c_str(),tokens[1].c_str(),tokens[1].c_str(),tokens[2].c_str(),tokens[3].c_str(),maf);
      }
   }
   printf("Finished single variant meta analysis.\n\n");

   fclose(output);
   if(outvcf) fclose(vcfout);
}

void Meta::Run(GroupFromAnnotation & group)
{
   printf("Refining SNPs in group ...\n");
   //Save maf for each annotation group to be tested for reuse
   Vector * maf;
   Vector * stats;
   Vector * singlePvalue;
   Vector * singleEffSize;
   Matrix * cov;
   maf = new Vector [group.annoGroups.Length()];
   stats = new Vector [group.annoGroups.Length()];
   cov = new Matrix [group.annoGroups.Length()];
   singlePvalue = new Vector [group.annoGroups.Length()];
   singleEffSize = new Vector [group.annoGroups.Length()];

   for(int g=0;g<group.annoGroups.Length();g++)
   {
      IntArray del; //del has the SNPs to be deleted from SNPlist
      maf[g].Dimension(0);
      stats[g].Dimension(0);
      singlePvalue[g].Dimension(0);
      singleEffSize[g].Dimension(0);
      int count = group.SNPlist[g].Length();
      for(int m=0;m<count;m++)
      {
	 String newSNP;
	 bool flipStatus = false;
	 double af = SNPmaf.Double(group.SNPlist[g][m]);
	 double singleP = singleVarPvalue.Double(group.SNPlist[g][m]);
	 double singleEff = singleVarEff.Double(group.SNPlist[g][m]);
	 if(af==_NAN_)
	 {
	    flipStatus=true;
	    StringArray SNPflip;
	    SNPflip.AddTokens(group.SNPlist[g][m],":");
	    newSNP = SNPflip[0]+":"+SNPflip[1]+":"+SNPflip[3]+":"+SNPflip[2];
	    af = SNPmaf.Double(newSNP);
	    singleP = singleVarPvalue.Double(newSNP);
	    singleEff = singleVarEff.Double(newSNP);
	 }
	 if(af==_NAN_ || af ==0.0 || af > MAF_cutoff || singleP==_NAN_)
	 {
	    del.Push(m);
	    continue;
	 }
	 double tmp;
	 if(flipStatus)
	    tmp = SNPstat.Double(newSNP);
	 else
	    tmp = SNPstat.Double(group.SNPlist[g][m]);
	 stats[g].Push(tmp);
	 maf[g].Push(af);

	 singlePvalue[g].Push(singleP);
	 singleEffSize[g].Push(singleEff);
      }

      //delete the SNPs that are not genotyped in any of the studies, 
      //or not polymorphic, or SNPs with maf>cutoff.
      for(int i=del.Length()-1;i>=0;i--)
      {
	 group.SNPlist[g].Delete(del[i]);
	 group.SNPNoAllele[g].Delete(del[i]);
      }
      count = group.SNPlist[g].Length();
      //printf("there are %d SNPs in this group\n",count);
      cov[g].Dimension(count,count,0.0);
   }

   //loop through cov matrices of all studies and update cov
   for(int study=0;study<studies.Length();study++)
   {
      printf("Reading cov matrix from study %s ...\n",studies[study].c_str());
      StringIntHash markerPosHash;
      StringArray markersInWindow,markersCov;
      String filename = studies[study] + ".singlevar.cov.txt";
      FILE * covfile;
      covfile  = fopen(filename,"r");
      int SNP_col = 2;
      int cov_col = 3;

      if(covfile == NULL)
      {
	 filename = studies[study] + ".MetaCov.assoc";
	 covfile  = fopen(filename,"r");
	 SNP_col = 4;
	 cov_col = 5;
      }

      if(covfile == NULL)
	 error("ERROR! Cannot open file: %s!\n",filename.c_str());

      String buffer;
      StringArray tokens;
      int m=0;
      while (!feof(covfile))
      {
	 buffer.ReadLine(covfile);
	 if(buffer.FindChar('#')!=-1 || buffer.Find("CHROM")!=-1)
	    continue;
	 tokens.Clear();
	 tokens.AddTokens(buffer, "\t ");
	 if(tokens[0].Find("chr")!=-1)
	    tokens[0] = tokens[0].SubStr(3);
	 String SNP = tokens[0] + ":" + tokens[1];
	 //exclude variants that have different REF and ALT
	 if(SNPexclude.Integer(study+":"+SNP)!=-1)
	 {
	    continue;
	 }
	 m++;
	 markerPosHash.SetInteger(SNP,m);
	 markersInWindow.Push(tokens[SNP_col]);
	 markersCov.Push(tokens[cov_col]);
      }
      fclose(covfile);

      printf("Updating group stats ...\n");
      //update group statistics
      for(int g=0;g<group.annoGroups.Length();g++)
      {
	 int count = group.SNPlist[g].Length();
	 StringArray pos,chr;
	 for(int i=0;i<count;i++)
	 {
	    StringArray tmp;
	    tmp.AddTokens(group.SNPNoAllele[g][i],":");
	    chr.Push(tmp[0]);
	    pos.Push(tmp[1]);
	 } //now pos has all the positions of markers in group g.

	 for(int m=0;m<count;m++)
	 {
	    int loc = markerPosHash.Integer(group.SNPNoAllele[g][m]);
	    //If this SNP is not genotpyed in this study, then skip
	    if(loc==-1)
	       continue;

	    //If this SNP from this study should be excluded due to non consistent ref/alt allele
	    //then skip
	    int skip = SNPexclude.Integer(study+":"+group.SNPNoAllele[g][m]);
	    if(skip!=-1)
	       continue;

	    //if this marker is flipped then markers from the entire row 
	    //should have cov multiply by -1.0.
	    int flip = flipSNP.Integer(study+":"+group.SNPNoAllele[g][m]);
	    double multiplyFactor=1.0;
	    if(flip!=-1)
	    {
	       multiplyFactor=-1.0;
	    }

	    //read through markersInWindow and find the selected markers
	    StringArray markers,markerscov;
	    markers.AddTokens(markersInWindow[loc-1],",");
	    markerscov.AddTokens(markersCov[loc-1],",");
	    StringIntHash markersNearByHash;
	    for(int l=0;l<markers.Length();l++)
	       markersNearByHash.SetInteger(markers[l],l);

	    //fill in the mth row of cov
	    for(int s=m;s<count;s++)
	    {
	       int p = markersNearByHash.Integer(pos[s]);
	       if(p==-1)
		  continue;
	       else
	       {
		  String markerName = study+":"+chr[s]+":"+pos[s];
		  //if the marker in window is supposed to be excluded due to non-consistent ref/alt allele
		  //then skip
		  int skip = SNPexclude.Integer(markerName);
		  if(skip!=-1)
		  {
		     continue;
		  }
		  int flip = flipSNP.Integer(markerName);
		  double factor=1.0;
		  if(flip!=-1)
		     factor=-1.0;
		  //printf("cov=%g,N=%d\n",markerscov[p].AsDouble(),SampleSize[study]);
		  cov[g][m][s] += multiplyFactor*factor*markerscov[p].AsDouble()*SampleSize[study];
	       }
	    }
	 }
      }
   }

   for(int g=0;g<group.annoGroups.Length();g++)
   {
      for(int r=0;r<cov[g].rows;r++)
	 for(int c=r+1;c<cov[g].cols;c++)
	    cov[g][c][r] = cov[g][r][c];
   }

   //Calculate the final meta statistics and group results
   if(Burden)
   {
      printf("Starting burden tests ...\n");
      //calculate final results here
      FILE * output;
      String filename;
      if(prefix =="")
	 filename = "meta.burden.results";
      else if(prefix.Last()=='.' || prefix.Last()=='/')
	 filename = prefix +  "meta.burden.results";
      else
	 filename = prefix + ".meta.burden.results";
      output=fopen(filename,"wt");

      String method = "BURDEN_";
      method += MAF_cutoff;

      FILE * reportOutput;
      if(report)
      {
	 String reportFile;
	 if(prefix =="")
	    reportFile = "meta.tophits.burden.tbl";
	 else if(prefix.Last()=='.' || prefix.Last()=='/')
	    reportFile = prefix +  "meta.tophits.burden.tbl";
	 else
	    reportFile = prefix + ".meta.tophits.burden.tbl";
	 reportOutput=fopen(reportFile,"wt");
	 fprintf(reportOutput,"GENE\tMETHOD\tGENE_PVALUE\tMAF_CUTOFF\tACTUAL_CUTOFF\tVARS\tMAFS\tEFFSIZES\tPVALUES\n");
      }

      fprintf(output,"##Method=Burden\n");
      fprintf(output,"##STUDY_NUM=%d\n",studies.Length());
      fprintf(output,"##TotalSampleSize=%d\n",total_N);
      if(fullResult)
	 fprintf(output,"#GROUPNAME\tNUM_VAR\tVARs\tMAFs\tSINGLEVAR_EFFECTs\tSINGLEVAR_PVALUEs\tAVG_AF\tMIN_AF\tMAX_AF\tEFFECT_SIZE\tPVALUE\n");
      else
	 fprintf(output,"#GROUPNAME\tNUM_VAR\tVARs\tAVG_AF\tMIN_AF\tMAX_AF\tEFFECT_SIZE\tPVALUE\n");

      double numerator,denominator,chisq,pvalue;
      for(int g=0;g<group.annoGroups.Length();g++)
      {
	 //	 printf("%s has %d SNPs.\n",group.annoGroups[g].c_str(),maf[g].Length());
	 //if(maf[g].Length()==1)
	 //printf("the SNPs are: %s with maf %g\n",group.SNPlist[g][0].c_str(),maf[g][0]);
	 if(maf[g].Length()==0)
	 {
	    if(fullResult)
	       fprintf(output,"%s\t0\t-\t-\t-\t-\t0\t0\t0\t-\t-\n",group.annoGroups[g].c_str());
	    else
	       fprintf(output,"%s\t0\t-\t0\t0\t0\t-\t-\n",group.annoGroups[g].c_str());
	    continue;
	 }

	 double average_af = maf[g].Average();
	 double min_af = maf[g].Min();
	 double max_af = maf[g].Max();

	 String var;
	 for(int i=0;i<maf[g].Length()-1;i++)
	 {
	    var += group.SNPlist[g][i] + ";";
	 }
	 var += group.SNPlist[g][maf[g].Length()-1];

	 Vector weight;
	 weight.Dimension(maf[g].Length());

	 for(int w=0;w<weight.Length();w++)
	 {
	    weight[w] = 1.0;
	 }

	 numerator  = weight.InnerProduct(stats[g]);
	 Vector tmp;
	 tmp.Dimension(group.SNPlist[g].Length());

	 for(int i=0;i<tmp.Length();i++)
	 {
	    tmp[i] = weight.InnerProduct(cov[g][i]);
	 }
	 denominator = tmp.InnerProduct(weight);

	 if(denominator==0.0)
	 {
	    if(fullResult)
	    {
	       fprintf(output,"%s\t%d\t%s\t",group.annoGroups[g].c_str(),group.SNPlist[g].Length(),var.c_str());

	       for(int i=0;i<maf[g].Length()-1;i++)
		  fprintf(output,"%g,",maf[g][i]);
	       fprintf(output,"%g\t",maf[g][maf[g].Length()-1]);

	       for(int i=0;i<singleEffSize[g].Length()-1;i++)
		  fprintf(output,"%g,",singleEffSize[g][i]);
	       fprintf(output,"%g\t",singleEffSize[g][singleEffSize[g].Length()-1]);

	       for(int i=0;i<singlePvalue[g].Length()-1;i++)
		  fprintf(output,"%g,",singlePvalue[g][i]);
	       fprintf(output,"%g\t",singlePvalue[g][singlePvalue[g].Length()-1]);

	       fprintf(output,"%g\t%g\t%g\t-\t-\n",average_af,min_af,max_af);
	    }
	    else
	       fprintf(output,"%s\t%d\t%s\t%g\t%g\t%g\t-\t-\n",group.annoGroups[g].c_str(),group.SNPlist[g].Length(),var.c_str(),average_af,min_af,max_af);
	    continue;
	 }

	 chisq = numerator*numerator/denominator;
	 double effSize = numerator/denominator;
	 //printf("gene:%s, numerator: %g,denominator %g,chisq %g\n",group.annoGroups[g].c_str(),numerator,denominator,chisq);
	 pvalue = pchisq(chisq,1,0,0);
	 bool disect=false;
	 while(pvalue==0.0)
	 {
	    disect=true;
	    chisq *= 0.999;
	    pvalue = pchisq(chisq,1,0,0);
	 }

	 if(fullResult)
	 {
	    fprintf(output,"%s\t%d\t%s\t",group.annoGroups[g].c_str(),group.SNPlist[g].Length(),var.c_str());

	    for(int i=0;i<maf[g].Length()-1;i++)
	       fprintf(output,"%g,",maf[g][i]);
	    fprintf(output,"%g\t",maf[g][maf[g].Length()-1]);

	    for(int i=0;i<singleEffSize[g].Length()-1;i++)
	       fprintf(output,"%g,",singleEffSize[g][i]);
	    fprintf(output,"%g\t",singleEffSize[g][singleEffSize[g].Length()-1]);

	    for(int i=0;i<singlePvalue[g].Length()-1;i++)
	       fprintf(output,"%g,",singlePvalue[g][i]);
	    fprintf(output,"%g\t",singlePvalue[g][singlePvalue[g].Length()-1]);

	    if(disect)
	       fprintf(output,"%g\t%g\t%g\t%g\t<%g\n",average_af,min_af,max_af,effSize,pvalue);
	    else
	       fprintf(output,"%g\t%g\t%g\t%g\t%g\n",average_af,min_af,max_af,effSize,pvalue);
	 }
	 else
	 {
	    if(disect)
	       fprintf(output,"%s\t%d\t%s\t%g\t%g\t%g\t%g\t<%g\n",group.annoGroups[g].c_str(),group.SNPlist[g].Length(),var.c_str(),average_af,min_af,max_af,effSize,pvalue);
	    else
	       fprintf(output,"%s\t%d\t%s\t%g\t%g\t%g\t%g\t%g\n",group.annoGroups[g].c_str(),group.SNPlist[g].Length(),var.c_str(),average_af,min_af,max_af,effSize,pvalue);
	 }

	 //tabulate top results here
	 if(pvalue <report_pvalue_cutoff && report)
	 {
	    StringArray variants;
	    variants.AddTokens(var,";");
	    for(int v=0;v<maf[g].Length();v++)
	    {
	       fprintf(reportOutput,"%s\t%s\t%g\t%g\t%g\t%s\t%g\t%g\t%g\n",group.annoGroups[g].c_str(),method.c_str(),pvalue,MAF_cutoff,MAF_cutoff,variants[v].c_str(),maf[g][v],singleEffSize[g][v],singlePvalue[g][v]);
	    }
	 }
      }
      fclose(output);
      if(report)
	 fclose(reportOutput);
      printf("Finished burden tests! \n\n");
   }

   if(MB)
   {
      printf("Starting MB tests ...\n");
      //calculate final results here
      FILE * output;
      String filename;
      if(prefix =="")
	 filename = "meta.MB.results";
      else if(prefix.Last()=='.' || prefix.Last()=='/')
	 filename = prefix +  "meta.MB.results";
      else
	 filename = prefix + ".meta.MB.results";
      output=fopen(filename,"wt");

      String method = "MB_";
      method += MAF_cutoff;

      FILE * reportOutput;
      if(report)
      {
	 String reportFile;
	 if(prefix =="")
	    reportFile = "meta.tophits.MB.tbl";
	 else if(prefix.Last()=='.' || prefix.Last()=='/')
	    reportFile = prefix +  "meta.tophits.MB.tbl";
	 else
	    reportFile = prefix + ".meta.tophits.MB.tbl";
	 reportOutput=fopen(reportFile,"wt");
	 fprintf(reportOutput,"GENE\tMETHOD\tGENE_PVALUE\tMAF_CUTOFF\tACTUAL_CUTOFF\tVARS\tMAFS\tEFFSIZES\tPVALUES\n");
      }

      fprintf(output,"##Method=MB\n");
      fprintf(output,"##STUDY_NUM=%d\n",studies.Length());
      fprintf(output,"##TotalSampleSize=%d\n",total_N);
      if(fullResult)
	 fprintf(output,"#GROUPNAME\tNUM_VAR\tVARs\tMAFs\tSINGLEVAR_EFFECTs\tSINGLEVAR_PVALUEs\tAVG_AF\tMIN_AF\tMAX_AF\tEFFECT_SIZE\tPVALUE\n");
      else
	 fprintf(output,"#GROUPNAME\tNUM_VAR\tVARs\tAVG_AF\tMIN_AF\tMAX_AF\tEFFECT_SIZE\tPVALUE\n");

      double numerator,denominator,chisq,pvalue;
      for(int g=0;g<group.annoGroups.Length();g++)
      {
	 //printf("%s has %d SNPs.\n",group.annoGroups[g].c_str(),maf[g].Length());
	 if(maf[g].Length()==0)
	 {
	    if(fullResult)
	       fprintf(output,"%s\t0\t-\t-\t-\t-\t0\t0\t0\t-\t-\n",group.annoGroups[g].c_str());
	    else
	       fprintf(output,"%s\t0\t-\t0\t0\t0\t-\t-\n",group.annoGroups[g].c_str());
	    continue;
	 }

	 double average_af = maf[g].Average();
	 double min_af = maf[g].Min();
	 double max_af = maf[g].Max();

	 String var;
	 for(int i=0;i<maf[g].Length()-1;i++)
	 {
	    var += group.SNPlist[g][i] + ";";
	 }
	 var += group.SNPlist[g][maf[g].Length()-1];

	 Vector weight;
	 weight.Dimension(maf[g].Length());

	 for(int w=0;w<weight.Length();w++)
	 {
	    weight[w] = 1.0/sqrt(maf[g][w]*(1.0-maf[g][w]));
	 }

	 numerator  = weight.InnerProduct(stats[g]);
	 Vector tmp;
	 tmp.Dimension(group.SNPlist[g].Length());

	 for(int i=0;i<tmp.Length();i++)
	 {
	    tmp[i] = weight.InnerProduct(cov[g][i]);
	 }
	 denominator = tmp.InnerProduct(weight);
	 if(denominator==0.0)
	 {
	    if(fullResult)
	    {
	       fprintf(output,"%s\t%d\t%s\t",group.annoGroups[g].c_str(),group.SNPlist[g].Length(),var.c_str());

	       for(int i=0;i<maf[g].Length()-1;i++)
		  fprintf(output,"%g,",maf[g][i]);
	       fprintf(output,"%g\t",maf[g][maf[g].Length()-1]);

	       for(int i=0;i<singleEffSize[g].Length()-1;i++)
		  fprintf(output,"%g,",singleEffSize[g][i]);
	       fprintf(output,"%g\t",singleEffSize[g][singleEffSize[g].Length()-1]);

	       for(int i=0;i<singlePvalue[g].Length()-1;i++)
		  fprintf(output,"%g,",singlePvalue[g][i]);
	       fprintf(output,"%g\t",singlePvalue[g][singlePvalue[g].Length()-1]);

	       fprintf(output,"%g\t%g\t%g\t-\t-\n",average_af,min_af,max_af);
	    }
	    else
	       fprintf(output,"%s\t%d\t%s\t%g\t%g\t%g\t-\t-\n",group.annoGroups[g].c_str(),group.SNPlist[g].Length(),var.c_str(),average_af,min_af,max_af);

	    continue;
	 }
	 chisq = numerator*numerator/denominator;
	 double effSize = numerator/denominator;
	 //	 printf("numerator: %g,denominator %g,chisq %g\n",numerator,denominator,chisq);
	 pvalue = pchisq(chisq,1,0,0);
	 bool disect=false;
	 while(pvalue==0.0)
	 {
	    disect=true;
	    chisq *= 0.999;
	    pvalue = pchisq(chisq,1,0,0);
	 }
	 //tabulate top results here
	 if(pvalue <report_pvalue_cutoff && report)
	 {
	    StringArray variants;
	    variants.AddTokens(var,";");
	    for(int v=0;v<maf[g].Length();v++)
	    {
	       fprintf(reportOutput,"%s\t%s\t%g\t%g\t%g\t%s\t%g\t%g\t%g\n",group.annoGroups[g].c_str(),method.c_str(),pvalue,MAF_cutoff,MAF_cutoff,variants[v].c_str(),maf[g][v],singleEffSize[g][v],singlePvalue[g][v]);
	    }
	 }

	 if(fullResult)
	 {
	    fprintf(output,"%s\t%d\t%s\t",group.annoGroups[g].c_str(),group.SNPlist[g].Length(),var.c_str());

	    for(int i=0;i<maf[g].Length()-1;i++)
	       fprintf(output,"%g,",maf[g][i]);
	    fprintf(output,"%g\t",maf[g][maf[g].Length()-1]);

	    for(int i=0;i<singleEffSize[g].Length()-1;i++)
	       fprintf(output,"%g,",singleEffSize[g][i]);
	    fprintf(output,"%g\t",singleEffSize[g][singleEffSize[g].Length()-1]);

	    for(int i=0;i<singlePvalue[g].Length()-1;i++)
	       fprintf(output,"%g,",singlePvalue[g][i]);
	    fprintf(output,"%g\t",singlePvalue[g][singlePvalue[g].Length()-1]);

	    if(disect)
	       fprintf(output,"%g\t%g\t%g\t%g\t<%g\n",average_af,min_af,max_af,effSize,pvalue);
	    else
	       fprintf(output,"%g\t%g\t%g\t%g\t%g\n",average_af,min_af,max_af,effSize,pvalue);
	 }
	 else
	 {
	    if(disect)
	       fprintf(output,"%s\t%d\t%s\t%g\t%g\t%g\t%g\t<%g\n",group.annoGroups[g].c_str(),group.SNPlist[g].Length(),var.c_str(),average_af,min_af,max_af,effSize,pvalue);
	    else
	       fprintf(output,"%s\t%d\t%s\t%g\t%g\t%g\t%g\t%g\n",group.annoGroups[g].c_str(),group.SNPlist[g].Length(),var.c_str(),average_af,min_af,max_af,effSize,pvalue);
	 }
      }
      fclose(output);
      if(report)
	 fclose(reportOutput);
      printf("Finished MB tests! \n\n");
   }

   if(VTa)
   {
      printf("\nStarting Variable Threshold tests ...\n");
      //calculate final results here
      FILE * output;
      String filename;
      if(prefix =="")
	 filename = "meta.VT.results";
      else if(prefix.Last()=='.' || prefix.Last()=='/')
	 filename = prefix +  "meta.VT.results";
      else
	 filename = prefix + ".meta.VT.results";
      output=fopen(filename,"wt");

      String method = "VT_";
      method += MAF_cutoff;
      FILE * reportOutput;
      if(report)
      {
	 String reportFile;
	 if(prefix =="")
	    reportFile = "meta.tophits.VT.tbl";
	 else if(prefix.Last()=='.' || prefix.Last()=='/')
	    reportFile = prefix +  "meta.tophits.VT.tbl";
	 else
	    reportFile = prefix + ".meta.tophits.VT.tbl";
	 reportOutput=fopen(reportFile,"wt");
	 fprintf(reportOutput,"GENE\tMETHOD\tGENE_PVALUE\tMAF_CUTOFF\tACTUAL_CUTOFF\tVARS\tMAFS\tEFFSIZES\tPVALUES\n");
      }

      fprintf(output,"##Method=VT\n");
      fprintf(output,"##STUDY_NUM=%d\n",studies.Length());
      fprintf(output,"##TotalSampleSize=%d\n",total_N);
      if(fullResult)
	 fprintf(output,"#GROUPNAME\tNUM_VAR\tVARs\tMAFs\tSINGLEVAR_EFFECTs\tSINGLEVAR_PVALUEs\tAVG_AF\tMIN_AF\tMAX_AF\tEFFECT_SIZE\tMAF_CUTOFF\tPVALUE\n");
      else
	 fprintf(output,"#GROUPNAME\tNUM_VAR\tVARs\tAVG_AF\tMIN_AF\tMAX_AF\tEFFECT_SIZE\tMAF_CUTOFF\tPVALUE\n");

      for(int g=0;g<group.annoGroups.Length();g++)
      {
	 if(g>1 && g%1000==1)
	    printf("Finished analyzing %d genes.\n",g-1);

	 if(maf[g].Length()==0)
	 {
	    if(fullResult)
	       fprintf(output,"%s\t0\t-\t-\t-\t-\t0\t0\t0\t-\t-\t-\n",group.annoGroups[g].c_str());
	    else
	       fprintf(output,"%s\t0\t-\t0\t0\t0\t-\t-\t-\n",group.annoGroups[g].c_str());
	    continue;
	 }

	 if(maf[g].Length()==1)
	 {
	    if(fullResult)
	    {
	       fprintf(output,"%s\t1\t%s\t%g\t%g\t%g\t%g\t%g\t%g\t%g\t%g\t%g\n",group.annoGroups[g].c_str(),group.SNPlist[g][0].c_str(),maf[g][0],singleEffSize[g][0],singlePvalue[g][0],maf[g][0],maf[g][0],maf[g][0],singleEffSize[g][0],maf[g][0],singlePvalue[g][0]);
	    }
	    else
	    {
	       fprintf(output,"%s\t1\t%s\t%g\t%g\t%g\t%g\t%g\t%g\n",group.annoGroups[g].c_str(),group.SNPlist[g][0].c_str(),maf[g][0],maf[g][0],maf[g][0],singleEffSize[g][0],maf[g][0],singlePvalue[g][0]);
	    }
	    continue;
	 }

	 //STEP1: sort maf[g] to find # of cutoffs
	 Vector cp_maf;
	 cp_maf.Copy(maf[g]);
	 cp_maf.Sort();
	 Vector maf_cutoff;
	 maf_cutoff.Push(cp_maf[0]);

	 for(int i=1;i<cp_maf.Length();i++)
	 {
	    if(cp_maf[i]>maf_cutoff[maf_cutoff.Length()-1])
	    {
	       maf_cutoff.Push(cp_maf[i]);	
	    }
	 } //now unique maf cutoffs are saved in maf_cutoff.

	 //STEP2: do burden test for each grouping method and find t_max
	 double pvalue,chosen_cutoff,chosen_effSize;
	 double numerator,denominator,t_max=_NAN_;
	 Vector weight,tmp,chosen_weight,score;
	 Matrix cov_weight;
	 weight.Dimension(maf[g].Length());
	 tmp.Dimension(group.SNPlist[g].Length());
	 cov_weight.Dimension(maf_cutoff.Length(),maf[g].Length());
	 score.Dimension(maf_cutoff.Length());

	 for(int i=0;i<maf_cutoff.Length();i++)
	 {
	    for(int w=0;w<weight.Length();w++)
	    {
	       if(maf[g][w]<=maf_cutoff[i])
		  weight[w]=1.0;
	       else
		  weight[w]=0.0;
	       cov_weight[i][w]=weight[w];
	    }
	    numerator = weight.InnerProduct(stats[g]);

	    for(int d=0;d<tmp.Length();d++)
	    {
	       tmp[d] = weight.InnerProduct(cov[g][d]);
	    }
	    denominator = tmp.InnerProduct(weight);

	    if(denominator != 0)
	    {
	       double t_stat = fabs(numerator/sqrt(denominator));
	       score[i]=t_stat;
	       if(t_max==_NAN_)
	       {
		  t_max = t_stat;
		  chosen_cutoff = maf_cutoff[i];
		  chosen_weight.Copy(weight);
		  chosen_effSize = numerator/denominator;
	       }
	       else
	       {
		  if(t_stat>t_max) 
		  {
		     t_max = t_stat;
		     chosen_cutoff = maf_cutoff[i];
		     chosen_weight.Copy(weight);
		     chosen_effSize = numerator/denominator;
		  }
	       }
	    }
	    else 
	       score[i]=0.0;
	 }

	 if(score.Max()==0.0)
	 {
	    if(fullResult)
	       fprintf(output,"%s\t-\t-\t-\t-\t-\t-\t-\t-\t-\t-\t-\n",group.annoGroups[g].c_str());
	    else
	       fprintf(output,"%s\t-\t-\t-\t-\t-\t-\t-\t-\n",group.annoGroups[g].c_str());
	    continue;
	 }
	 /*
	    if(chosen_weight.Sum()==1.0)
	    {
	    int index;
	    for(int i=0;i<chosen_weight.Length();i++)
	    {
	    if(chosen_weight[i]==1.0)
	    {
	    index=i;
	    break;
	    }
	    }

	    if(fullResult)
	    {
	    fprintf(output,"%s\t1\t%s\t%g\t%g\t%g\t%g\t%g\t%g\t%g\t%g\t%g\n",group.annoGroups[g].c_str(),group.SNPlist[g][index].c_str(),maf[g][index],singleEffSize[g][index],singlePvalue[g][index],maf[g][index],maf[g][index],maf[g][index],singleEffSize[g][index],maf[g][index],singlePvalue[g][index]);
	    }
	    else
	    {
	    fprintf(output,"%s\t1\t%s\t%g\t%g\t%g\t%g\t%g\t%g\n",group.annoGroups[g].c_str(),group.SNPlist[g][index].c_str(),maf[g][index],maf[g][index],maf[g][index],singleEffSize[g][index],maf[g][index],singlePvalue[g][index]);
	    }
	    continue;
	    }
	  */
	 Vector tmp_maf,tmp_eff,tmp_pvalue;
	 for(int i=0;i<maf[g].Length();i++)
	 {
	    if(chosen_weight[i]==1.0)
	    {
	       tmp_maf.Push(maf[g][i]);
	    }
	 }

	 for(int i=0;i<maf[g].Length();i++)
	 {
	    if(chosen_weight[i]==1.0)
	    {
	       tmp_eff.Push(singleEffSize[g][i]);
	       tmp_pvalue.Push(singlePvalue[g][i]);
	    }
	 }

	 double average_af = tmp_maf.Average();
	 double min_af = tmp_maf.Min();
	 double max_af = tmp_maf.Max();

	 String var;
	 for(int i=0;i<maf[g].Length()-1;i++)
	 {
	    if(chosen_weight[i]==1.0)
	       var += group.SNPlist[g][i] + ";";
	 }
	 if(chosen_weight[maf[g].Length()-1]==1.0)
	    var += group.SNPlist[g][maf[g].Length()-1];

	 //STEP3: calculate covariance matrix for (U_1 ... U_#cutoff) 
	 Matrix cov_U,cov_U_tmp;
	 cov_U_tmp.Product(cov_weight,cov[g]);
	 Matrix cov_weight_trans;
	 cov_weight_trans.Transpose(cov_weight);
	 cov_U.Product(cov_U_tmp,cov_weight_trans); //now, cov(U) is saved in cov_U
	 /*
	    printf("cov_U is:\n");
	    for(int row=0;row<cov_U.rows;row++)
	    {
	    for(int col=0;col<cov_U.cols;col++)
	    {
	    printf("%g\t",cov_U[row][col]);
	    }
	    printf("\n");
	    }
	  */
	 //Calculate covariance matrix for (T_1 ... T_#cutoff)
	 Matrix cov_T;
	 cov_T.Dimension(cov_U.rows,cov_U.cols);
	 cov2cor(cov_U,cov_T);

	 //STEP4: calculate VT pvalue and report.
	 int cutoff = maf_cutoff.Length();
	 double * lower = new double [cutoff];
	 double * upper = new double [cutoff];
	 double * mean = new double [cutoff];

	 for(int i=0;i<cutoff;i++)
	 {
	    mean[i] = 0.0;
	    lower[i] = -t_max;
	    upper[i] = t_max;
	 }

	 //Use pmvnorm to calculate the asymptotic p-value
	 Vector result;
	 pmvnorm(lower,upper,mean,cov_T,false,result);
	 //printf("pmvnorm results:%g,%g,%g\n",result[0],result[1],result[2]);
	 /*	    printf("maf_cutoff is:\n");
		    for(int item=0;item<maf_cutoff.Length();item++)
		    printf("%g\t",maf_cutoff[item]);
		    printf("\n");

		    printf("maf[g] is:\n");
		    for(int item=0;item<maf[g].Length();item++)
		    printf("%g\t",maf[g][item]);
		    printf("\n");

		    printf("cov_weight is:\n");
		    for(int row=0;row<cov_weight.rows;row++)
		    {
		    for(int col=0;col<cov_weight.cols;col++)
		    {
		    printf("%g\t",cov_weight[row][col]);
		    }
		    printf("\n");
		    }


		    printf("cov[g] is:\n");
		    for(int row=0;row<cov[g].rows;row++)
		    {
		    for(int col=0;col<cov[g].cols;col++)
		    {
		    printf("%g\t",cov[g][row][col]);
		    }
		    printf("\n");
		    }
		    printf("cov_U is:\n");
		    for(int row=0;row<cov_T.rows;row++)
		    {
		    for(int col=0;col<cov_T.cols;col++)
		    {
		    printf("%g\t",cov_U[row][col]);
		    }
		    printf("\n");
		    }

		    printf("cov_T is:\n");
		    for(int row=0;row<cov_T.rows;row++)
		    {
		    for(int col=0;col<cov_T.cols;col++)
		    {
		    printf("%g\t",cov_T[row][col]);
		    }
		    printf("\n");
		    }
	  */
	 /*
	    if(result[0]==-1.0)
	    {
	    SVD svd;
	    svd.Decompose(cov_T);
	    Matrix corr;

	 //reconstruct corr by eigen values and eigen vectors.
	 Matrix tmp,trans_v;
	 tmp.Dimension(svd.w.Length(),svd.w.Length());
	 for(int w=0;w<svd.w.Length();w++)
	 {
	 if(svd.w[w]<0) 
	 {
	 svd.w[w] =0.0;
	 }
	 }
	 for(int i=0;i<tmp.rows;i++)
	 for(int j=0;j<tmp.cols;j++)
	 tmp[i][j]=svd.u[i][j]*svd.w[j];
	 trans_v.Transpose(svd.v);
	 corr.Product(tmp,trans_v);
	 printf("new cov_T is:\n");
	 for(int row=0;row<cov_T.rows;row++)
	 {          
	 for(int col=0;col<cov_T.cols;col++)
	 {       
	 printf("%g\t",corr[row][col]);
	 }       
	 printf("\n");
	 }
	 result.Copy(pmvnorm(lower,upper,mean,corr,false));
	 }
	  */
	 if(result[0]==-1.0)
	 {
	    if(fullResult)
	    {
	       fprintf(output,"%s\t%d\t%s\t",group.annoGroups[g].c_str(),tmp_maf.Length(),var.c_str());

	       for(int i=0;i<tmp_maf.Length()-1;i++)
		  fprintf(output,"%g,",tmp_maf[i]);
	       fprintf(output,"%g\t",tmp_maf[tmp_maf.Length()-1]);

	       for(int i=0;i<tmp_eff.Length()-1;i++)
		  fprintf(output,"%g,",tmp_eff[i]);
	       fprintf(output,"%g\t",tmp_eff[tmp_eff.Length()-1]);

	       for(int i=0;i<tmp_pvalue.Length()-1;i++)
		  fprintf(output,"%g,",tmp_pvalue[i]);
	       fprintf(output,"%g\t",tmp_pvalue[tmp_pvalue.Length()-1]);

	       fprintf(output,"%g\t%g\t%g\t%g\t%g\tERROR:CORR_NOT_POS_SEMI_DEF\n",average_af,min_af,max_af,chosen_effSize,chosen_cutoff);
	    }
	    else
	       fprintf(output,"%s\t%d\t%s\t%g\t%g\t%g\t%g\t%g\tERROR:CORR_NOT_POS_SEMI_DEF\n",group.annoGroups[g].c_str(),tmp_maf.Length(),var.c_str(),average_af,min_af,max_af,chosen_effSize,chosen_cutoff);
	 }
	 else
	 {
	    if(1.0-result[0]==0.0)
	    {
	       //	    printf("gene %s has result %g\n",group.annoGroups[g].c_str(),1.0-result[0]);
	       printf("Using Shuang's algorithm to calculate MVN pvalue for gene %s ... ",group.annoGroups[g].c_str());
	       if(maf_cutoff.Length()>20)
	       {
		  while(maf_cutoff.Length()>20)
		  {
		     maf_cutoff.Delete(0);
		  }

		  double numerator,denominator,t_max=_NAN_;
		  Vector weight,tmp,chosen_weight,score;
		  Matrix cov_weight;
		  weight.Dimension(maf[g].Length());
		  tmp.Dimension(group.SNPlist[g].Length());
		  cov_weight.Dimension(maf_cutoff.Length(),maf[g].Length());

		  for(int i=0;i<maf_cutoff.Length();i++)
		  {
		     for(int w=0;w<weight.Length();w++)
		     {
			if(maf[g][w]<=maf_cutoff[i])
			   weight[w]=1.0;
			else
			   weight[w]=0.0;
			cov_weight[i][w]=weight[w];
		     }
		     numerator = weight.InnerProduct(stats[g]);

		     for(int d=0;d<tmp.Length();d++)
		     {
			tmp[d] = weight.InnerProduct(cov[g][d]);
		     }
		     denominator = tmp.InnerProduct(weight);
		     if(denominator != 0)
		     {
			double t_stat = fabs(numerator/sqrt(denominator));
			score.Push(t_stat);
			if(t_max==_NAN_)
			{
			   t_max = t_stat;
			   chosen_cutoff = maf_cutoff[i];
			   chosen_weight.Copy(weight);
			   chosen_effSize = numerator/denominator;
			}
			else
			{
			   if(t_stat>t_max)
			   {
			      t_max = t_stat;
			      chosen_cutoff = maf_cutoff[i];
			      chosen_weight.Copy(weight);
			      chosen_effSize = numerator/denominator;
			   }
			}
		     }
		     else
			score.Push(0.0);
		  }
		  if(score.Max()==0.0)
		  {
		     if(fullResult)
			fprintf(output,"%s\t-\t-\t-\t-\t-\t-\t-\t-\t-\t-\t-\n",group.annoGroups[g].c_str());
		     else
			fprintf(output,"%s\t-\t-\t-\t-\t-\t-\t-\t-\n",group.annoGroups[g].c_str());
		     continue;
		     printf("completed!\n");
		  }

		  Vector tmp_maf,tmp_eff,tmp_pvalue;
		  for(int i=0;i<maf[g].Length();i++)
		  {
		     if(chosen_weight[i]==1.0)
		     {
			tmp_maf.Push(maf[g][i]);
		     }
		  }

		  for(int i=0;i<maf[g].Length();i++)
		  {
		     if(chosen_weight[i]==1.0)
		     {
			tmp_eff.Push(singleEffSize[g][i]);
			tmp_pvalue.Push(singlePvalue[g][i]);
		     }
		  }
		  average_af = tmp_maf.Average();
		  min_af = tmp_maf.Min();
		  max_af = tmp_maf.Max();

		  String var;
		  for(int i=0;i<maf[g].Length()-1;i++)
		  {
		     if(chosen_weight[i]==1.0)
			var += group.SNPlist[g][i] + ";";
		  }
		  if(chosen_weight[maf[g].Length()-1]==1.0)
		     var += group.SNPlist[g][maf[g].Length()-1];

		  //STEP3: calculate covariance matrix for (U_1 ... U_#cutoff) 
		  Matrix cov_U,cov_U_tmp;
		  cov_U_tmp.Product(cov_weight,cov[g]);
		  Matrix cov_weight_trans;
		  cov_weight_trans.Transpose(cov_weight);
		  cov_U.Product(cov_U_tmp,cov_weight_trans); //now, cov(U) is saved in cov_U
		  //Calculate covariance matrix for (T_1 ... T_#cutoff)
		  Matrix cov_T;
		  cov_T.Dimension(cov_U.rows,cov_U.cols);
		  cov2cor(cov_U,cov_T);

		  pvalue = CalculateMVTPvalue(score,cov_T,t_max);
		  printf("completed!\n");
	       }
	       else
	       {
		  pvalue = CalculateMVTPvalue(score,cov_T,t_max);
		  printf("completed!\n");
	       }
	    }
	    else
	    {
	       pvalue = 1.0 - result[0];
	    }

	    if(pvalue <report_pvalue_cutoff && report)
	    {
	       StringArray variants;
	       variants.AddTokens(var,";");
	       for(int v=0;v<tmp_maf.Length();v++)
	       {
		  fprintf(reportOutput,"%s\t%s\t%g\t%g\t%g\t%s\t%g\t%g\t%g\n",group.annoGroups[g].c_str(),method.c_str(),pvalue,MAF_cutoff,chosen_cutoff,variants[v].c_str(),tmp_maf[v],tmp_eff[v],tmp_pvalue[v]);
	       }
	    }

	    if(fullResult)
	    {
	       fprintf(output,"%s\t%d\t%s\t",group.annoGroups[g].c_str(),tmp_maf.Length(),var.c_str());

	       for(int i=0;i<tmp_maf.Length()-1;i++)
		  fprintf(output,"%g,",tmp_maf[i]);
	       fprintf(output,"%g\t",tmp_maf[tmp_maf.Length()-1]);

	       for(int i=0;i<tmp_eff.Length()-1;i++)
		  fprintf(output,"%g,",tmp_eff[i]);
	       fprintf(output,"%g\t",tmp_eff[tmp_eff.Length()-1]);

	       for(int i=0;i<tmp_pvalue.Length()-1;i++)
		  fprintf(output,"%g,",tmp_pvalue[i]);
	       fprintf(output,"%g\t",tmp_pvalue[tmp_pvalue.Length()-1]);

	       fprintf(output,"%g\t%g\t%g\t%g\t%g\t%g\n",average_af,min_af,max_af,chosen_effSize,chosen_cutoff,pvalue);
	    }
	    else
	       fprintf(output,"%s\t%d\t%s\t%g\t%g\t%g\t%g\t%g\t%g\n",group.annoGroups[g].c_str(),tmp_maf.Length(),var.c_str(),average_af,min_af,max_af,chosen_effSize,chosen_cutoff,pvalue);
	 }

	 if(lower) delete [] lower;
	 if(upper) delete [] upper;
	 if(mean) delete [] mean;
      }
      fclose(output);
      if(report)
	 fclose(reportOutput);
      printf("Finished Variable Threshold tests.\n\n");
   }

   if(SKAT)
   {
      printf("Starting working on SKAT ...\n");
      //calculate Q statistics here
      FILE * output;
      String filename;
      if(prefix =="")
	 filename = "meta.SKAT.results";
      else if(prefix.Last()=='.' || prefix.Last()=='/')
	 filename = prefix +  "meta.SKAT.results";
      else
	 filename = prefix + ".meta.SKAT.results";
      output=fopen(filename,"wt");

      String method = "SKAT_";
      method += MAF_cutoff;
      FILE * reportOutput;
      if(report)
      {
	 String reportFile;
	 if(prefix =="")
	    reportFile = "meta.tophits.SKAT.tbl";
	 else if(prefix.Last()=='.' || prefix.Last()=='/')
	    reportFile = prefix +  "meta.tophits.SKAT.tbl";
	 else
	    reportFile = prefix + ".meta.tophits.SKAT.tbl";
	 reportOutput=fopen(reportFile,"wt");
	 fprintf(reportOutput,"GENE\tMETHOD\tGENE_PVALUE_DAVIES\tGENE_PVALUE_LIU\tMAF_CUTOFF\tACTUAL_CUTOFF\tVAR\tMAF\tEFFSIZE\tPVALUE\n");
      }

      fprintf(output,"##Method=SKAT\n");
      fprintf(output,"##STUDY_NUM=%d\n",studies.Length());
      fprintf(output,"##TotalSampleSize=%d\n",total_N);
      if(fullResult)
	 fprintf(output,"#GROUPNAME\tNUM_VAR\tVARs\tMAFs\tSINGLEVAR_EFFECTs\tSINGLEVAR_PVALUEs\tAVG_AF\tMIN_AF\tMAX_AF\tSTATISTICS\tPVALUE_DAVIES\tPVALUE_LIU\n");
      else
	 fprintf(output,"#GROUPNAME\tNUM_VAR\tVARs\tAVG_AF\tMIN_AF\tMAX_AF\tSTATISTICS\tPVALUE_DAVIES\tPVALUE_LIU\n");

      double Qstat,pvalue,pvalue_liu;
      SVD svd;
      Matrix Pmix;
      for(int g=0;g<group.annoGroups.Length();g++)
      {
	 //      printf("Now working on group %s\n",group.annoGroups[g].c_str());
	 if(g%1000==1 && g>1000)
	 {
	    printf("Finished analyzing %d genes.\n",((int)g/1000)*1000);
	 }
	 Vector weight;
	 int n = maf[g].Length();
	 if(maf[g].Length()==0)
	 {
	    if(fullResult)
	       fprintf(output,"%s\t0\t-\t-\t-\t-\t0\t0\t0\t-\t-\t-\n",group.annoGroups[g].c_str());
	    else
	       fprintf(output,"%s\t0\t-\t0\t0\t0\t-\t-\t-\n",group.annoGroups[g].c_str());
	    continue;
	 }

	 double average_af = maf[g].Average();
	 double min_af = maf[g].Min();
	 double max_af = maf[g].Max();

	 String var;
	 for(int i=0;i<maf[g].Length()-1;i++)
	 {
	    var += group.SNPlist[g][i] + ";";
	 }
	 var += group.SNPlist[g][maf[g].Length()-1];

	 //refine variants  
	 Vector cp_maf;
	 cp_maf.Copy(maf[g]);

	 if(n>1)
	 {
	    //calculate correlation of cov[g]_i and cov[g]_j
	    IntArray del;
	    double corr;
	    for(int i=0;i<n;i++)
	    {
	       for(int j=i+1;j<n;j++)
	       {
		  corr = CalculateCorrCoef(cov[g][i],cov[g][j]);
		  if(corr ==1.0 || corr == -1.0)
		  {
		     del.Push(i);
		     break;
		  }
	       }
	    }
	    //del.RemoveDuplicates();

	    for(int i=del.Length()-1;i>=0;i--)
	    {
	       cov[g].DeleteRow(del[i]);
	       cov[g].DeleteColumn(del[i]);
	       stats[g].Delete(del[i]);
	       cp_maf.Delete(del[i]);
	       n--;
	    }
	 }
	 weight.Dimension(n);
	 //get weight based on maf
	 double alpha = 1.0;
	 double beta=25.0;
	 double beta_density;
	 for(int w=0;w<n;w++)
	 {
	    beta_density = GetBetaDensity(alpha,beta,cp_maf[w]);
	    weight[w] = beta_density * beta_density;
	 }
	 Vector tmp;
	 tmp.Dimension(n);
	 for(int i=0;i<n;i++)
	 {
	    tmp[i] = weight[i]*stats[g][i];
	 }
	 Qstat = tmp.InnerProduct(stats[g]);
	 //printf("Qstat is: %g\n",Qstat);
	 svd.Decompose(cov[g]);
	 Matrix L;
	 L.Dimension(n,n);
	 for(int j=0;j<n;j++)
	 {
	    double sqrt_w = sqrt(fabs(svd.w[j]));
	    for(int i=0;i<n;i++)
	    {
	       L[i][j] = svd.u[i][j] * sqrt_w;
	    }
	 }

	 Matrix Pmix_tmp,Pmix;
	 Pmix.Dimension(n,n);
	 Pmix_tmp.Dimension(n,n);
	 for(int j=0;j<n;j++)
	 {
	    for(int i=0;i<n;i++)
	    {
	       Pmix_tmp[i][j] = L[i][j] * weight[j];
	    }
	 }
	 Matrix L_trans;
	 L_trans.Transpose(L);
	 Pmix.Product(Pmix_tmp,L_trans);
	 svd.Decompose(Pmix);

	 double * lambda = new double [n];
	 for(int i=0;i<n;i++)
	 {
	    lambda[i] = fabs(svd.w[i]);
	 }
	 double Q_stat = Qstat;
	 double Qstat_liu = Qstat;

	 pvalue = MixChidist(lambda, n, Qstat,"Davies");

	 bool disect_davies=false,disect_liu = false;
	 int disect_itr=0;
	 while( (pvalue<=0.0 ||pvalue==2.0 || isnan(pvalue)) && disect_itr<10000)
	 {
	    disect_davies=true;
	    Qstat*=0.9999;
	    pvalue = MixChidist(lambda, n, Qstat,"Davies");
	    disect_itr++;
	 }
	 disect_itr=0;
	 while((pvalue<=0.0 ||pvalue==2.0 || isnan(pvalue))  && disect_itr<10000)
	 {
	    Qstat*=0.99;
	    pvalue = MixChidist(lambda, n, Qstat,"Davies");
	    disect_itr++;
	 }
	 pvalue_liu = MixChidist(lambda, n, Q_stat,"Liu");
	 disect_itr=0;
	 while( (pvalue_liu<=0.0 ||pvalue_liu==2.0 || isnan(pvalue_liu)) && disect_itr<10000)
	 {
	    disect_liu=true;
	    Qstat_liu*=0.9999;
	    pvalue_liu = MixChidist(lambda, n, Qstat_liu,"Liu");
	    disect_itr++;
	 }
	 disect_itr=0;
	 while((pvalue_liu<=0.0 ||pvalue_liu==2.0 || isnan(pvalue_liu))  && disect_itr<10000)
	 {
	    Qstat_liu*=0.99;
	    pvalue_liu = MixChidist(lambda, n, Qstat_liu,"Liu");
	    disect_itr++;
	 }
	 if(isnan(pvalue_liu) || isnan(pvalue))
	 {
	    if(fullResult)
	    {
	       fprintf(output,"%s\t%d\t%s\t",group.annoGroups[g].c_str(),group.SNPlist[g].Length(),var.c_str());

	       for(int i=0;i<maf[g].Length()-1;i++)
		  fprintf(output,"%g,",maf[g][i]);
	       fprintf(output,"%g\t",maf[g][maf[g].Length()-1]);

	       for(int i=0;i<singleEffSize[g].Length()-1;i++)
		  fprintf(output,"%g,",singleEffSize[g][i]);
	       fprintf(output,"%g\t",singleEffSize[g][singleEffSize[g].Length()-1]);

	       for(int i=0;i<singlePvalue[g].Length()-1;i++)
		  fprintf(output,"%g,",singlePvalue[g][i]);
	       fprintf(output,"%g\t",singlePvalue[g][singlePvalue[g].Length()-1]);

	       fprintf(output,"%g\t%g\t%g\t%g\t-\t-\n",average_af,min_af,max_af,Q_stat);
	    }
	    else
	       fprintf(output,"%s\t%d\t%s\t%g\t%g\t%g\t%g\t-\t-\n",group.annoGroups[g].c_str(),group.SNPlist[g].Length(),var.c_str(),average_af,min_af,max_af,Q_stat);
	    continue;
	 }
	 //tabulate top results
	 if((pvalue <report_pvalue_cutoff || pvalue_liu<report_pvalue_cutoff)  && report)
	 {
	    StringArray variants;
	    variants.AddTokens(var,";");
	    for(int v=0;v<maf[g].Length();v++)
	    {
	       fprintf(reportOutput,"%s\t%s\t%s%g\t%s%g\t%g\t%g\t%s\t%g\t%g\t%g\n",group.annoGroups[g].c_str(),method.c_str(),disect_davies?"<":"",pvalue,disect_liu?"<":"",pvalue_liu,MAF_cutoff,MAF_cutoff,variants[v].c_str(),maf[g][v],singleEffSize[g][v],singlePvalue[g][v]);
	    }
	 }

	 if(fullResult)
	 {
	    fprintf(output,"%s\t%d\t%s\t",group.annoGroups[g].c_str(),group.SNPlist[g].Length(),var.c_str());

	    for(int i=0;i<maf[g].Length()-1;i++)
	       fprintf(output,"%g,",maf[g][i]);
	    fprintf(output,"%g\t",maf[g][maf[g].Length()-1]);

	    for(int i=0;i<singleEffSize[g].Length()-1;i++)
	       fprintf(output,"%g,",singleEffSize[g][i]);
	    fprintf(output,"%g\t",singleEffSize[g][singleEffSize[g].Length()-1]);

	    for(int i=0;i<singlePvalue[g].Length()-1;i++)
	       fprintf(output,"%g,",singlePvalue[g][i]);
	    fprintf(output,"%g\t",singlePvalue[g][singlePvalue[g].Length()-1]);

	    fprintf(output,"%g\t%g\t%g\t%g\t%s%g\t%s%g\n",average_af,min_af,max_af,Q_stat,disect_davies?"<":"",pvalue,disect_liu?"<":"",pvalue_liu);
	 }
	 else
	 {
	    fprintf(output,"%s\t%d\t%s\t%g\t%g\t%g\t%g\t%s%g\t%s%g\n",group.annoGroups[g].c_str(),group.SNPlist[g].Length(),var.c_str(),average_af,min_af,max_af,Q_stat,disect_davies?"<":"",pvalue,disect_liu?"<":"",pvalue_liu);
	 }
	 if(lambda) delete [] lambda;
      }
      fclose(output);
      if(report)
	 fclose(reportOutput);
      printf("Finished working on SKAT.\n\n");
   }

   //housekeeping
   if(maf) delete [] maf;
   if(stats) delete [] stats;
   if(cov) delete [] cov;
   if(singlePvalue) delete [] singlePvalue;
   if(singleEffSize) delete [] singleEffSize;
}

double Meta::GetBetaDensity(double a, double b, double x)
{
   double density;
   density = exp(gammln(a+b)-gammln(a)-gammln(b)+(a-1.0)*log(x)+(b-1.0)*log(1.0-x));
   return density;
}

double Meta::CalculateCorrCoef(Vector & a,Vector & b)
{
   double coef;
   int n = a.Length();
   double sum_a = a.Sum();
   double sum_b = b.Sum();
   coef =  n*a.InnerProduct(b) - sum_a*sum_b;
   coef /= sqrt((n*a.SumSquares()-sum_a*sum_a)*(n*b.SumSquares()-sum_b*sum_b));  
   return coef;
}



